import 'package:apptive_grid_core/apptive_grid_core.dart';
import 'package:flutter_test/flutter_test.dart';

void main() {
  group('String', () {
    test('Value is set', () {
      final entity = StringDataEntity('Value');

      expect(entity.value, 'Value');
      expect(entity.schemaValue, 'Value');
    });

    test('Default is null', () {
      final entity = StringDataEntity();

      expect(entity.value, null);
    });
  });

  group('Integer', () {
    test('Value is set', () {
      final entity = IntegerDataEntity(3);

      expect(entity.value, 3);
      expect(entity.schemaValue, 3);
    });

    test('Default is null', () {
      final entity = StringDataEntity();

      expect(entity.value, null);
    });
  });

  group('Decimal', () {
    test('Value is set', () {
      final entity = DecimalDataEntity(47.11);

      expect(entity.value, 47.11);
      expect(entity.schemaValue, 47.11);
    });

    test('Default is null', () {
      final entity = StringDataEntity();

      expect(entity.value, null);
    });
  });

  group('Date', () {
    test('Value is set', () {
      final date = DateTime(
        2020,
        3,
        3,
      );
      final entity = DateDataEntity(date);

      expect(entity.value, date);
      expect(entity.schemaValue, '2020-03-03');
    });

    test('Json is parsed', () {
      final date = DateTime(
        2020,
        3,
        3,
      );
      final entity = DateDataEntity.fromJson('2020-03-03');

      expect(entity.value, date);
      expect(entity.schemaValue, '2020-03-03');
    });

    test('Default is null', () {
      final entity = DateDataEntity();

      expect(entity.value, null);
    });
  });

  group('DateTime', () {
    test('Value is set', () {
      final date = DateTime(2020, 3, 3, 12, 12, 12);
      final entity = DateTimeDataEntity(date);

      expect(entity.value, date);
      expect(entity.schemaValue, '2020-03-03T11:12:12.000Z');
    });

    test('Json is parsed', () {
      final date = DateTime(2020, 3, 3, 12, 12, 12);
      final entity = DateTimeDataEntity.fromJson('2020-03-03T11:12:12.000Z');

      expect(entity.value, date);
      expect(entity.schemaValue, '2020-03-03T11:12:12.000Z');
    });

    test('Default is null', () {
      final entity = DateTimeDataEntity();

      expect(entity.value, null);
    });
  });

  group('Boolean', () {
    test('Value is set', () {
      final entity = BooleanDataEntity(true);

      expect(entity.value, true);
      expect(entity.schemaValue, true);
    });

    test('Default is false', () {
      final entity = BooleanDataEntity();

      expect(entity.value, false);
    });
  });

  group('Enum', () {
    test('Value is set', () {
      const value = 'value';
      final values = {'value', 'otherValue'};
      final entity = EnumDataEntity(value: value, options: values);

      expect(entity.value, value);
      expect(entity.options, values);
      expect(entity.schemaValue, value);
    });

    test('Default is null', () {
      final entity = EnumDataEntity();

      expect(entity.value, null);
      expect(entity.options, []);
    });
  });

  group('EnumCollection', () {
    test('Value is set', () {
      const value = 'value';
      final values = {'value', 'otherValue'};
      final entity = EnumCollectionDataEntity(value: {value}, options: values);

      expect(entity.value, {value});
      expect(entity.options, values);
      expect(entity.schemaValue, [value]);
    });

    test('Default is empty list', () {
      final entity = EnumCollectionDataEntity();

      expect(entity.value, <String>{});
      expect(entity.options, <String>{});
      expect(entity.schemaValue, null);
    });

    group('Equality', () {
      test('Equals', () {
        final a =
            EnumCollectionDataEntity(value: {'A', 'B'}, options: {'A', 'B'});
        final b =
            EnumCollectionDataEntity(value: {'A', 'B'}, options: {'A', 'B'});

        expect(a, equals(b));
        expect(a.hashCode, equals(b.hashCode));
      });

      test(
          'Different Order Value Set '
          'equals but different Hashcode', () {
        final a =
            EnumCollectionDataEntity(value: {'A', 'B'}, options: {'A', 'B'});
        final b =
            EnumCollectionDataEntity(value: {'B', 'A'}, options: {'A', 'B'});

        expect(a, equals(b));
        expect(a.hashCode, isNot(equals(b.hashCode)));
      });

      test('Unequals', () {
        final a =
            EnumCollectionDataEntity(value: {'A', 'B'}, options: {'A', 'B'});
        final b = EnumCollectionDataEntity(
          value: {
            'B',
          },
          options: {'A', 'B'},
        );

        expect(a, isNot(equals(b)));
        expect(a.hashCode, isNot(equals(b.hashCode)));
      });
    });
  });

  group('Equality', () {
    final string = StringDataEntity('Value');
    final stringEquals = StringDataEntity('Value');
    final stringUnequals = StringDataEntity('otherValue');
    final integer = IntegerDataEntity(2);
    final date = DateDataEntity.fromJson('2020-03-03');
    final dateTime = DateTimeDataEntity.fromJson('2020-03-03T12:12:12.000');
    final boolean = BooleanDataEntity(true);
    final selection =
        EnumDataEntity(value: 'value', options: {'value', 'otherValue'});
    final equalSelection =
        EnumDataEntity(value: 'value', options: {'value', 'otherValue'});
    final unEqualSelection =
        EnumDataEntity(value: 'otherValue', options: {'value', 'otherValue'});

    test('equals', () {
      expect(string == stringEquals, true);
      expect(string.hashCode - stringEquals.hashCode == 0, true);
      expect(selection == equalSelection, true);
      expect(selection.hashCode - equalSelection.hashCode == 0, true);
    });
    test('not equals', () {
      expect(string == stringUnequals, false);
      expect(string.hashCode - stringUnequals.hashCode == 0, false);
      expect(string.hashCode != integer.hashCode, true);
      expect(string.hashCode != date.hashCode, true);
      expect(string.hashCode != dateTime.hashCode, true);
      expect(string.hashCode != boolean.hashCode, true);
      expect(selection == unEqualSelection, false);
      expect(selection.hashCode - unEqualSelection.hashCode == 0, false);
    });
  });

  group('Attachment', () {
    test('Value is set', () {
      final attachments = <Attachment>[
        Attachment(name: 'one', url: Uri.parse('one'), type: 'image/png'),
        Attachment(name: 'two', url: Uri.parse('two'), type: 'image/png'),
      ];
      final entity = AttachmentDataEntity(attachments);

      expect(entity.value, attachments);
      expect(
        entity.schemaValue,
        attachments.map((attachment) => attachment.toJson()),
      );
    });

    test('From Json', () {
      final attachments = <Attachment>[
        Attachment(name: 'one', url: Uri.parse('one'), type: 'image/png'),
        Attachment(name: 'two', url: Uri.parse('two'), type: 'image/png'),
      ];
      final direct = AttachmentDataEntity(attachments);

      final fromJson = AttachmentDataEntity.fromJson(
        attachments.map((attachment) => attachment.toJson()).toList(),
      );

      expect(direct, equals(fromJson));
      expect(direct.hashCode, equals(fromJson.hashCode));
    });

    test('Null json is Empty List', () {
      final entity = AttachmentDataEntity.fromJson(null);

      expect(entity.value, equals([]));
      expect(entity.schemaValue, isNull);
    });

    test('Default is Empty List', () {
      final entity = AttachmentDataEntity();

      expect(entity.value, equals([]));
      expect(entity.schemaValue, isNull);
    });
  });
}
